package com.gci.cdenv.data.consume.app.modules.common.fence;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * @author lyz
 * @date 2023/9/1 11:58
 */
public class GeoFenceChecker {


    // 用于处理浮点数比较误差
    private static final double EPSILON = 1e-10;

    public static boolean isInsidePolygon(List<Coordinate> polygonCoordinates, Coordinate point) {
        List<Coordinate> newList = removeDuplicates(polygonCoordinates);
        int intersectCount = 0;

        for (int i = 0; i < newList.size(); i++) {
            Coordinate p1 = newList.get(i);
            Coordinate p2 = newList.get((i + 1) % newList.size());
            // 检查点是否在多边形的边上
            if (isPointOnSegment(p1, p2, point)) {
                // 如果点在边上，立即返回true
                return true;
            }
            double angle = calculateAngle(p1, point, p2);
            // 检查点是否在多边形的顶点上
            if (angle == 0 && pointIsOnSegment(p1, p2, point)) {
                // 如果点在顶点上，立即返回true
                return true;
            }
            // 检查点是否与边相交
            if (angle > 0 && !pointIsOnSegment(p1, p2, point)) {
                intersectCount++;
            }
        }
        return intersectCount % 2 != 0;
    }

    /**
     * 检查点是否与边相交
     * @param p1
     * @param p2
     * @param point
     * @return
     */
    private static boolean isPointOnSegment(Coordinate p1, Coordinate p2, Coordinate point) {
        return pointIsOnSegment(p1, p2, point) && isPointBetween(p1, p2, point);
    }

    /**
     * 检查点是否在边上并且不与边的端点重合
     * @param p1
     * @param p2
     * @param point
     * @return
     */
    private static boolean pointIsOnSegment(Coordinate p1, Coordinate p2, Coordinate point) {
        double crossProduct = calculateCrossProduct(p1, p2, point);
        return Math.abs(crossProduct) < EPSILON && isPointBetween(p1, p2, point);
    }

    /**
     * 检查点是否在边的投影范围内
     * @param p1
     * @param p2
     * @param point
     * @return
     */
    private static boolean isPointBetween(Coordinate p1, Coordinate p2, Coordinate point) {
        return point.getLongitude() >= Math.min(p1.getLongitude(), p2.getLongitude()) &&
                point.getLongitude() <= Math.max(p1.getLongitude(), p2.getLongitude()) &&
                point.getLatitude() >= Math.min(p1.getLatitude(), p2.getLatitude()) &&
                point.getLatitude() <= Math.max(p1.getLatitude(), p2.getLatitude());
    }

    private static double calculateAngle(Coordinate p1, Coordinate p2, Coordinate point) {
        double angle1 = Math.atan2(p1.getLatitude() - point.getLatitude(), p1.getLongitude() - point.getLongitude());
        double angle2 = Math.atan2(p2.getLatitude() - point.getLatitude(), p2.getLongitude() - point.getLongitude());

        double angle = angle1 - angle2;
        if (angle < 0) {
            angle += 2 * Math.PI;
        }

        return angle;
    }

    private static double calculateCrossProduct(Coordinate p1, Coordinate p2, Coordinate point) {
        return (p1.getLongitude() - point.getLongitude()) * (p2.getLatitude() - point.getLatitude()) -
                (p1.getLatitude() - point.getLatitude()) * (p2.getLongitude() - point.getLongitude());
    }

    private static List<Coordinate> removeDuplicates(List<Coordinate> list) {
        Set<Coordinate> set = new HashSet<>(list);
        return new ArrayList<>(set);
    }


    public static void main(String[] args) {
        List<Coordinate> list = new ArrayList<>();
        Coordinate d1 = new Coordinate(104.028826,30.571496);

        Coordinate d2 = new Coordinate(105.10682,30.568777);

        Coordinate d3 = new Coordinate(104.132072,30.398718);

        Coordinate d4 = new Coordinate(104.113488,30.34468);

        Coordinate d5 = new Coordinate(104.113488,30.34468);

        Coordinate d6 = new Coordinate(104.114283,30.341593);

        Coordinate d7 = new Coordinate(104.005321,30.367152);

        Coordinate d8 = new Coordinate(103.908782,30.412965);


        list.add(d6);
        list.add(d7);
        list.add(d8);
        list.add(d1);
        list.add(d2);
        list.add(d3);
        list.add(d4);
        list.add(d5);

        Coordinate d = new Coordinate(104.101848,30.445424);
        long start = System.currentTimeMillis();
        if (GeoFenceChecker.isInsidePolygon(list,d)) {
            System.out.println("点在围栏内部");
        } else {
            System.out.println("点在围栏外部");
        }
        long end = System.currentTimeMillis();
        System.out.println("计算结束："+(end-start)+"ms");
    }
}
