package com.gci.cdenv.data.consume.app.modules.consumer;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.util.NumberUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import com.alibaba.fastjson.JSONObject;
import com.gci.cdenv.data.consume.app.modules.common.RedisKey;
import com.gci.cdenv.data.consume.app.modules.gps.dto.*;
import com.gci.cdenv.data.consume.app.modules.gps.service.GisActiveSafetyService;
import com.gci.cdenv.data.consume.app.modules.gps.service.GpsConsumerService;
import com.gci.cdenv.data.consume.app.modules.gps.service.VehicleThreeElectricService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.ApplicationArguments;
import org.springframework.boot.ApplicationRunner;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.scheduling.concurrent.ThreadPoolTaskExecutor;
import org.springframework.stereotype.Component;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.stream.Collectors;

/**
 * @author lyz
 * @date 2023/9/8 20:13
 */
@Component
@Slf4j
public class QueueConsumer implements ApplicationRunner {

    @Autowired
    private StringRedisTemplate redisTemplate;

    @Autowired
    private StringRedisTemplate redisTemplate2;

    @Autowired
    private VehicleThreeElectricService vehicleThreeElectricService;

    @Autowired
    private GisActiveSafetyService gisActiveSafetyService;

    // 定义定时休眠的时间（毫秒）
    private static final long SLEEP_TIME_MILLIS = 5 * 1000;

    @Autowired
    private ThreadPoolTaskExecutor taskExecutor;

    @Autowired
    private GpsConsumerService gpsConsumerService;

    private static final int EIC_QUEUE_NUMBER = 0;
    private static final int GPS_QUEUE_NUMBER = 1;
    private static final int DSM_QUEUE_NUMBER = 2;
    private static final int ADAS_QUEUE_NUMBER = 3;

    @Value("${dataApp.lineCount}")
    private Integer lineCount;


    private BlockingQueue<List<String>> eicQueue = new LinkedBlockingQueue<>();
    private BlockingQueue<List<String>> gpsQueue = new LinkedBlockingQueue<>();
    private BlockingQueue<List<String>> dsmQueue = new LinkedBlockingQueue<>();
    private BlockingQueue<List<String>> adasQueue = new LinkedBlockingQueue<>();


    @Override
    public void run(ApplicationArguments args) throws Exception {
        for (int queueNumber = 0; queueNumber < 4; queueNumber++) {
            final int finalQueueNumber = queueNumber;
            taskExecutor.execute(() -> consumeQueue(finalQueueNumber));
        }
        // 启动四个线程分别从四个不同的Redis队列中获取数据并放入对应的BlockingQueue
        taskExecutor.execute(() -> fetchDataFromRedis(RedisKey.EIC_QEUENAME, eicQueue));
        taskExecutor.execute(() -> fetchDataFromRedis(RedisKey.GPS_QEUENAME, gpsQueue));
        taskExecutor.execute(() -> fetchDataFromRedis(RedisKey.GCI_SB_DSM, dsmQueue));
        taskExecutor.execute(() -> fetchDataFromRedis(RedisKey.GCI_SB_ADAS, adasQueue));
    }

    private void fetchDataFromRedis(String queueName, BlockingQueue<List<String>> targetQueue) {
        while (true) {
            try {
                List<String> data = redisTemplate.opsForList().leftPop(queueName, lineCount);
                if (data != null && !data.isEmpty()) {
                    // 将数据放入对应的BlockingQueue
                    redisTemplate2.opsForList().rightPushAll(queueName,data);
                    targetQueue.offer(data);
                } else {
                    Thread.sleep(SLEEP_TIME_MILLIS);
                }
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            } catch (Exception e) {
                e.printStackTrace();
                log.error("从Redis队列获取数据时发生异常，队列名称：{}, 异常：{}", queueName, e.getMessage());
            }
        }
    }

    private void consumeQueue(int queueNumber) {
        BlockingQueue<List<String>> queue = getQueueByNumber(queueNumber);
        while (true) {
            try {
                // 阻塞等待数据
                List<String> dataList = queue.take();
                if (dataList != null && !dataList.isEmpty()) {
                    processQueueData(dataList, queueNumber);
                }
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            } catch (Exception e) {
                log.error("处理队列数据时发生异常，队列编号：{}, 异常：{}", queueNumber, e);
            }
        }
    }

    private BlockingQueue<List<String>> getQueueByNumber(int queueNumber) {
        switch (queueNumber) {
            case 0:
                return eicQueue;
            case 1:
                return gpsQueue;
            case 2:
                return dsmQueue;
            case 3:
                return adasQueue;
            default:
                throw new IllegalArgumentException("无效的队列编号: " + queueNumber);
        }
    }

    private void processQueueData(List<String> data, int queueNumber) {
//        logQueueData(queueNumber, data);
        try {
            switch (queueNumber) {
                case EIC_QUEUE_NUMBER:
                    processEicData(data);
                    break;
                case GPS_QUEUE_NUMBER:
                    processGpsData(data);
                    break;
                case DSM_QUEUE_NUMBER:
                    processActiveSafetyData(data);
                    break;
                case ADAS_QUEUE_NUMBER:
                    processActiveSafetyData(data);
                    break;
            }
        } catch (Exception e) {
            e.printStackTrace();
            log.error("处理数据时出错，队列编号：{}, 异常：{}", queueNumber, e.getMessage());
        }
    }

    private void logQueueData(int queueNumber, String data) {
        String queueName = getQueueName(queueNumber);
        log.info("获取{}数据{}", queueName, data);
    }

    private void processEicData(List<String> data) {
        List<VehicleThreeElectric> dtos = data.stream().map(d -> {
            VehicleThreeElectric vehicleThreeElectric = JSONObject.parseObject(d, VehicleThreeElectric.class);
            halfUp(vehicleThreeElectric);
            return vehicleThreeElectric;
        }).filter(d -> isValidEicData(d)).collect(Collectors.toList());
        vehicleThreeElectricService.handle(dtos);
        List<GpsDto> gpsDtos = vehicleThreeElectricService.buildGpsDto(dtos);
        gpsConsumerService.processLocationData(gpsDtos);
    }

    private void halfUp(VehicleThreeElectric vehicleThreeElectric) {
        VehicleDTO vehicle = vehicleThreeElectric.getVehicle();
        if (BeanUtil.isNotEmpty(vehicle)) {
            vehicle.setSpeed(vehicle.getSpeed().setScale(2, BigDecimal.ROUND_HALF_UP));
            vehicle.setTotalCurrent(vehicle.getTotalCurrent().setScale(2, BigDecimal.ROUND_HALF_UP));
            vehicle.setTotalMileage(vehicle.getTotalMileage().setScale(2, BigDecimal.ROUND_HALF_UP));
            vehicle.setTotalVoltage(vehicle.getTotalVoltage().setScale(2, BigDecimal.ROUND_HALF_UP));
        }

        List<ElectricMotorDTO> electricMotor = vehicleThreeElectric.getElectricMotor();
        if (ObjectUtil.isNotEmpty(electricMotor)) {
            for (ElectricMotorDTO electricMotorDTO : electricMotor) {
                if (StrUtil.isNotBlank(electricMotorDTO.getDcBusCurrent())) {
                    electricMotorDTO.setDcBusCurrent(
                            NumberUtil.toBigDecimal(electricMotorDTO.getDcBusCurrent())
                                    .setScale(2, BigDecimal.ROUND_HALF_UP).toString());
                }
                if (StrUtil.isNotBlank(electricMotorDTO.getInputVoltage())) {
                    electricMotorDTO.setInputVoltage(
                            NumberUtil.toBigDecimal(electricMotorDTO.getInputVoltage())
                                    .setScale(2, BigDecimal.ROUND_HALF_UP).toString()
                    );
                }
                if (StrUtil.isNotBlank(electricMotorDTO.getTemperature())) {
                    electricMotorDTO.setTemperature(
                            NumberUtil.toBigDecimal(electricMotorDTO.getTemperature())
                                    .setScale(2, BigDecimal.ROUND_HALF_UP).toString()
                    );
                }
                if (StrUtil.isNotBlank(electricMotorDTO.getTorque())) {
                    electricMotorDTO.setTorque(
                            NumberUtil.toBigDecimal(electricMotorDTO.getTorque())
                                    .setScale(2, BigDecimal.ROUND_HALF_UP).toString()
                    );
                }
            }
        }
        ExtremeDataDTO extremeData = vehicleThreeElectric.getExtremeData();
        if (BeanUtil.isNotEmpty(extremeData)) {
            if (StrUtil.isNotBlank(extremeData.getMaxCellVoltage())) {
                extremeData.setMaxCellVoltage(
                        NumberUtil.toBigDecimal(extremeData.getMaxCellVoltage())
                                .setScale(2, BigDecimal.ROUND_HALF_UP).toString()
                );
            }
            if (StrUtil.isNotBlank(extremeData.getMinCellVoltage())) {
                extremeData.setMinCellVoltage(
                        NumberUtil.toBigDecimal(extremeData.getMinCellVoltage())
                                .setScale(2, BigDecimal.ROUND_HALF_UP).toString()
                );
            }
        }
    }


    private void processGpsData(List<String> data) {
        List<GpsDto> dtos = data.stream().map(d -> {
            return JSONObject.parseObject(d, GpsDto.class);
        }).collect(Collectors.toList());
        gpsConsumerService.processLocationData(dtos);
    }

    private void processActiveSafetyData(List<String> data) {
        data.parallelStream().forEach(d -> {
            ActiveSafetyDto dto = JSONObject.parseObject(d, ActiveSafetyDto.class);
            gisActiveSafetyService.gisActiveSafety(dto);
        });
    }


    private boolean isValidEicData(VehicleThreeElectric eic) {
        if (Objects.isNull(eic.getVehicle())) {
            log.info("无效数据，整车数据没有，过滤eic记录,{}", JSONObject.toJSONString(eic));
            return false;
        }
        return true;
    }

    private String getQueueName(int queueNumber) {
        switch (queueNumber) {
            case EIC_QUEUE_NUMBER:
                return RedisKey.EIC_QEUENAME;
            case GPS_QUEUE_NUMBER:
                return RedisKey.GPS_QEUENAME;
            case DSM_QUEUE_NUMBER:
                return RedisKey.GCI_SB_DSM;
            case ADAS_QUEUE_NUMBER:
                return RedisKey.GCI_SB_ADAS;
            default:
                return "";
        }
    }
}
