package com.gci.cdenv.data.consume.app.modules.gps.service;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.json.JSONUtil;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.gci.cdenv.data.consume.app.modules.common.ConstantKey;
import com.gci.cdenv.data.consume.app.modules.common.RedisKey;
import com.gci.cdenv.data.consume.app.modules.common.fence.Coordinate;
import com.gci.cdenv.data.consume.app.modules.common.fence.GeoFenceChecker;
import com.gci.cdenv.data.consume.app.modules.common.util.CoordinateConverterUtil;
import com.gci.cdenv.data.consume.app.modules.common.util.JsonUtil;
import com.gci.cdenv.data.consume.app.modules.common.util.LatLonPoint;
import com.gci.cdenv.data.consume.app.modules.common.util.PrimaryKeyUtil;
import com.gci.cdenv.data.consume.app.modules.gps.dto.*;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @author lyz
 * @date 2023/8/31 17:44
 */
@Service
@RequiredArgsConstructor
@Slf4j
public class GisFenceService {

    @Autowired
    private  StringRedisTemplate redisTemplate;

    @Autowired
    private VehicleBindDriverService vehicleBindDriverService;

    /**
     * 处理围栏计算，并返回是否在围栏外，默认false
     * @param vehicleInfoDTO
     * @param gpsDto
     * @return
     */
    public boolean gisFenceAndBackOutFence(VehicleInfoDTO vehicleInfoDTO,GpsDto gpsDto) {
        //没有gps坐标信息，不参与围栏计算
        if(ObjectUtil.isEmpty(gpsDto.getData().getLat())||ObjectUtil.isEmpty(gpsDto.getData().getLon())){
//            log.error("围栏计算失败，无经纬度数据。按否处理,数据-{}",JSONObject.toJSONString(gpsDto));
            return false;
        }
        boolean currentFlag = false;
        try{
            //获取车牌号
            String numberPlate = vehicleInfoDTO.getPlateNumber();
            //获取车辆对应的围栏信息
            List<FenceDTO> fenceDTOS = Optional.ofNullable(redisTemplate.opsForValue().get(RedisKey.VEHICLE_FENCE_KEY+numberPlate)).map(a->{
                return JSONUtil.toList(StrUtil.removeAll(StrUtil.strip(a, "\""),"\\"), FenceDTO.class).stream().filter(b->{
                    return b.isWithinFenceTime(gpsDto.getData().getGpstime());
                }).collect(Collectors.toList());
            }).orElse(new ArrayList<>());
            //转换高德信息
            LatLonPoint point = CoordinateConverterUtil.toWGSPoint(gpsDto.getData().getLat().doubleValue(),gpsDto.getData().getLon().doubleValue());
            Coordinate coordinate = new Coordinate(point.getLongitude(),point.getLatitude());
            //遍历车辆的围栏信息
            for(FenceDTO fence:fenceDTOS){
                //当前车辆的围栏多边形数据及当前车辆的坐标数据。
                List<Coordinate> polygon = fence.getCoordinateList();
                //判断本次是否在围栏内
                boolean isInFence = GeoFenceChecker.isInsidePolygon(polygon,coordinate);
                //连续在围栏内的次数
                int consecutiveInsideCount = Optional.ofNullable(redisTemplate.opsForValue().get(RedisKey.CONSECUTIVE_INSIDE_COUNT + numberPlate+fence.getFenceId()))
                        .map(Integer::parseInt).orElse(0);
                //连续在围栏外的次数
                int consecutiveOutsideCount = Optional.ofNullable(redisTemplate.opsForValue().get(RedisKey.CONSECUTIVE_OUT_COUNT + numberPlate+fence.getFenceId()))
                        .map(Integer::parseInt).orElse(0);
                //从redis获取最新状态是否在围栏内
                boolean isInsideLast = Optional.ofNullable(redisTemplate.opsForValue().get(RedisKey.INSIDE_FENCE + numberPlate+fence.getFenceId()))
                        .map(value -> value.equals("true")).orElse(true);
                if (!isInFence) {
                    //如果在场外 则清掉场内次数，并累加场外次数，超过6次，，则生成告警信息
                    consecutiveInsideCount = 0;
                    consecutiveOutsideCount++;
                    if (consecutiveOutsideCount >= ConstantKey.DEFAULT_FENCE_COUNT) {
                        // 最新状态设置为在围栏外
                        redisTemplate.opsForValue().set(RedisKey.INSIDE_FENCE + numberPlate+fence.getFenceId(), "false");
                        currentFlag = true;
                        //当前判断在场外，且最新状态在场内，则生成一次告警记录
                        if(isInsideLast){
                            //生成告警信息
                            generateFenceStartAlarm(vehicleInfoDTO,fence,gpsDto.getData(),point);
                        }
                    }
                } else {
                    //如果当前在场内，但是最新状态在场外,则场外次数清零。场内累加，超过6次则判断在围栏内
                    if(!isInsideLast){
                        consecutiveOutsideCount = 0;
                        consecutiveInsideCount++;
                        if (consecutiveInsideCount >= ConstantKey.DEFAULT_FENCE_COUNT) {
                            redisTemplate.delete(RedisKey.INSIDE_FENCE + numberPlate+fence.getFenceId());
                            //输出围栏进场，完善围栏告警完结信息
                            generateFenceEndAlarm(vehicleInfoDTO,fence,gpsDto.getData(),point);
                        }
                    }
                }
                redisTemplate.opsForValue().set(RedisKey.CONSECUTIVE_INSIDE_COUNT + numberPlate+fence.getFenceId(), String.valueOf(consecutiveInsideCount));
                redisTemplate.opsForValue().set(RedisKey.CONSECUTIVE_OUT_COUNT + numberPlate+fence.getFenceId(), String.valueOf(consecutiveOutsideCount));
            }
        }catch (Exception e){
            e.printStackTrace();
            log.error("围栏计算失败，按否处理,数据-{}",JSONObject.toJSONString(gpsDto));
        }
        return currentFlag;
    }

    /**
     * 生成告警信息，并推送至redis队列
     */
    private void generateFenceStartAlarm(VehicleInfoDTO vehicle, FenceDTO fenceDTO, GpsData gpsData,LatLonPoint point) {
        FenceAlarm fenceAlarm = new FenceAlarm();
        fenceAlarm.setId(PrimaryKeyUtil.gengerateId());
        fenceAlarm.setProjectName(vehicle.getProjectName());
        fenceAlarm.setFenceId(fenceDTO.getFenceId());
        fenceAlarm.setProjectId(vehicle.getProjectId());

        DriverSignInDto driverDto =  vehicleBindDriverService.getCacheDriverInfo(vehicle.getPlateNumber());
        if(Objects.nonNull(driverDto)){
            fenceAlarm.setDriverId(driverDto.getDriverId());
            fenceAlarm.setDriverName(driverDto.getDriverName());
        }
        fenceAlarm.setFenceType(fenceDTO.getFenceType());
        fenceAlarm.setName(fenceDTO.getFenceName());
        fenceAlarm.setPlateNumber(vehicle.getPlateNumber());
        fenceAlarm.setAlarmTime(gpsData.getGpstime());
        fenceAlarm.setLon(gpsData.getLon());
        fenceAlarm.setLat(gpsData.getLat());
        fenceAlarm.setGdLon(new BigDecimal(point.getLongitude()));
        fenceAlarm.setGdLat(new BigDecimal(point.getLatitude()));

        Alarm alarm = new Alarm();
        //1-围栏 2-超速
        alarm.setType(1);
        // 1-开始  2-结束
        alarm.setFlag(1);
        alarm.setTime(new Date());
        alarm.setData(JSONObject.parseObject(JSONObject.toJSONString(fenceAlarm)));
        // TODO: 2023/10/18 屏蔽
        //redisTemplate.opsForZSet().add(RedisKey.FENCE_VEHICLE_SCORE,vehicle.getPlateNumber()+"-"+fenceAlarm.getId()+"-"+fenceDTO.getFenceId(), fenceDTO.estimateTime(gpsData.getGpstime()));
        //推送至告警信息队列
        redisTemplate.opsForList().leftPush(RedisKey.FENCE_SPEED_ALARM_BEAN, JSONUtil.toJsonPrettyStr(alarm));
    }

    /**
     * 生成告警信息，并推送至redis队列
     */
    private void generateFenceEndAlarm(VehicleInfoDTO vehicle, FenceDTO fenceDTO, GpsData gpsData,LatLonPoint point) {
        FenceAlarm fenceAlarm = new FenceAlarm();
        fenceAlarm.setProjectName(vehicle.getProjectName());
        fenceAlarm.setFenceId(fenceDTO.getFenceId());
        fenceAlarm.setProjectId(vehicle.getProjectId());

        fenceAlarm.setFenceType(fenceDTO.getFenceType());
        fenceAlarm.setName(fenceDTO.getFenceName());
        fenceAlarm.setPlateNumber(vehicle.getPlateNumber());

        DriverSignInDto driverDto =  vehicleBindDriverService.getCacheDriverInfo(vehicle.getPlateNumber());
        if(Objects.nonNull(driverDto)){
            fenceAlarm.setDriverId(driverDto.getDriverId());
            fenceAlarm.setDriverName(driverDto.getDriverName());
        }
        fenceAlarm.setAlarmTime(gpsData.getGpstime());
        fenceAlarm.setLon(gpsData.getLon());
        fenceAlarm.setLat(gpsData.getLat());
        fenceAlarm.setGdLon(new BigDecimal(point.getLongitude()));
        fenceAlarm.setGdLat(new BigDecimal(point.getLatitude()));

        Alarm alarm = new Alarm();
        //1-围栏 2-超速
        alarm.setType(1);
        // 1-开始  2-结束
        alarm.setFlag(2);
        alarm.setTime(new Date());
        alarm.setData(JSONObject.parseObject(JSONObject.toJSONString(fenceAlarm)));
        redisTemplate.opsForList().leftPush(RedisKey.FENCE_SPEED_ALARM_BEAN,JSONUtil.toJsonPrettyStr(alarm));
    }
}
