package com.gci.cdenv.data.consume.app.modules.gps.service;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.StrUtil;
import cn.hutool.json.JSONUtil;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.gci.cdenv.data.consume.app.modules.common.ConstantKey;
import com.gci.cdenv.data.consume.app.modules.common.RedisKey;
import com.gci.cdenv.data.consume.app.modules.common.util.CoordinateConverterUtil;
import com.gci.cdenv.data.consume.app.modules.common.util.LatLonPoint;
import com.gci.cdenv.data.consume.app.modules.gps.entity.GisSpeedingThreshold;
import com.gci.cdenv.data.consume.app.modules.gps.mapper.GisSpeedingThresholdMapper;
import com.gci.cdenv.data.consume.app.modules.gps.vo.SpeedRedisVO;
import com.gci.cdenv.data.consume.app.modules.gps.dto.*;
import com.gci.cdenv.data.consume.app.modules.gps.entity.GisSpeedingAlarm;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.util.*;

/**
 * 超速处理
 */
@Service
@Slf4j
@RequiredArgsConstructor
public class GisSpeedService {

    @Value("${param.standSpeed}")
    private Integer standSpeed;
    @Value("${param.speedCount}")
    private Integer speedCount;

    private static final Integer MAX_SPEED = 220;

    @Autowired
    private  StringRedisTemplate redisTemplate;

    @Autowired
    private VehicleBindDriverService vehicleBindDriverService;

    @Autowired
    private GisSpeedingThresholdMapper gisSpeedingThresholdMapper;

    /**
     * 超速计算，生成超速记录，并返回当前车辆是否超速标志位
     * @param vehicleInfoDTO
     * @param gpsDto
     * @return
     */
    public boolean gisSpeedAndBacIsSpeed(VehicleInfoDTO vehicleInfoDTO, GpsDto gpsDto) {
        //获取车牌号
        String numberPlate = vehicleInfoDTO.getPlateNumber();
        //判断本次是否超速
        GisSpeedingThreshold speedingThreshold = checkSpeed(gpsDto.getData().getSpeed(),vehicleInfoDTO);
        Boolean isSpeed = speedingThreshold.getIsSpeed();
        //获取连续超速次数及首次超速时间，首次超速经纬度，速度
        String redisStr = redisTemplate.opsForValue().get(RedisKey.GIS_CONSECUTIVE_SPEED_COUNT + numberPlate);
        SpeedRedisVO speedRedis = null;
        if(StrUtil.isNotBlank(redisStr)){
            speedRedis = JSONUtil.toBean(redisStr, SpeedRedisVO.class);
        }
        if (isSpeed) {
            //当前超速，更新redis最新超速次数（时间为首次超速插入时间）。
            if(Objects.isNull(speedRedis)){
                SpeedRedisVO vo = new SpeedRedisVO();
                vo.setCount(1);
                vo.setStartTime(DateUtil.format(gpsDto.getData().getGpstime(),"yyyy-MM-dd HH:mm:ss"));
                vo.setStartLon(gpsDto.getData().getLon());
                vo.setStartLat(gpsDto.getData().getLat());
                List<Double> speedList = new ArrayList<>();
                speedList.add(gpsDto.getData().getSpeed().setScale(2,BigDecimal.ROUND_HALF_UP).doubleValue());
                vo.setSpeedList(speedList);
                redisTemplate.opsForValue().set(RedisKey.GIS_CONSECUTIVE_SPEED_COUNT + numberPlate,JSONUtil.toJsonPrettyStr(vo));
                //首次超速，推送一条超速开始的到业务系统
                Alarm alarm = buildBean(gpsDto.getData(),vehicleInfoDTO,speedRedis,1,speedingThreshold);
                redisTemplate.opsForList().leftPush(RedisKey.FENCE_SPEED_ALARM_BEAN, JSONUtil.toJsonPrettyStr(alarm));
            }else{
                //开始时间不超过结束时间，忽略中途插入的数据
                if(DateUtil.parse(speedRedis.getStartTime(),"yyyy-MM-dd HH:mm:ss").isBefore(gpsDto.getData().getGpstime())){
                    //连续超速，记录多次速度
                    speedRedis.setCount(speedRedis.getCount() + 1);
                    List<Double> speedList = speedRedis.getSpeedList();
                    speedList.add(gpsDto.getData().getSpeed().setScale(2,BigDecimal.ROUND_HALF_UP).doubleValue());
                    speedRedis.setSpeedList(speedList);
                    redisTemplate.opsForValue().set(RedisKey.GIS_CONSECUTIVE_SPEED_COUNT + numberPlate,JSONUtil.toJsonPrettyStr(speedRedis));
                }
            }
        }else {
            if(Objects.nonNull(speedRedis)){
                //开始时间不超过结束时间
                if(DateUtil.parse(speedRedis.getStartTime(),"yyyy-MM-dd HH:mm:ss").isBefore(gpsDto.getData().getGpstime())){
                    //当前不超速，判断连续超速次数是否达到阈值。达到阈值则先保存数据到数据库，否则清空redis该条数据
                    if(speedRedis.getCount() >= speedCount){
                        //超速结束，推送一条超速结束的到业务系统
                        Alarm alarm = buildBean(gpsDto.getData(),vehicleInfoDTO,speedRedis,2, speedingThreshold);
                        //推送
                        redisTemplate.opsForList().leftPush(RedisKey.FENCE_SPEED_ALARM_BEAN, JSONUtil.toJsonPrettyStr(alarm));
                        //清理redis数据
                        redisTemplate.delete(RedisKey.GIS_CONSECUTIVE_SPEED_COUNT + numberPlate);
                    }
                };
            }
        }
        return isSpeed;
    }

    //判断是否超速
    private GisSpeedingThreshold checkSpeed(BigDecimal speed, VehicleInfoDTO vehicleInfoDTO) {
        GisSpeedingThreshold target = null;
        //如果速度异常或无效，则返回false
        if(speed.intValue() == ConstantKey.INVALID || speed.intValue() == ConstantKey.EX){
            target = new GisSpeedingThreshold();
            target.setIsSpeed(false);
            return target;
        }
        if(speed.compareTo(BigDecimal.ZERO) < 0 || speed.compareTo(new BigDecimal(MAX_SPEED)) >= 0){
            target = new GisSpeedingThreshold();
            target.setIsSpeed(false);
            return target;
        }
        String s = redisTemplate.opsForValue().get(RedisKey.THRESHOLD + vehicleInfoDTO.getProjectId() + ":" + vehicleInfoDTO.getVehicleType());
        GisSpeedingThreshold speedingThreshold = JSONObject.parseObject(StrUtil.removeAll(StrUtil.strip(s, "\""),"\\"), GisSpeedingThreshold.class);
        if (BeanUtil.isNotEmpty(speedingThreshold)) {
            target = speedingThreshold;
        } else {
            log.error("超速阈值规则获取失败！检查redis数据!");
            List<GisSpeedingThreshold> gisSpeedingThresholds = gisSpeedingThresholdMapper.selectList(new LambdaUpdateWrapper<GisSpeedingThreshold>()
                    .eq(GisSpeedingThreshold::getProjectId, vehicleInfoDTO.getProjectId())
                    .eq(GisSpeedingThreshold::getIsValid,1)
                    .or().eq(GisSpeedingThreshold::getThresholdType, '0')
                    .orderByDesc(GisSpeedingThreshold::getThresholdType));
            for (GisSpeedingThreshold gisSpeedingThreshold : gisSpeedingThresholds) {
                //车辆种类>项目>通用
                if (vehicleInfoDTO.getProjectId().equals(gisSpeedingThreshold.getProjectId())
                        && vehicleInfoDTO.getVehicleType().equals(gisSpeedingThreshold.getVehicleType())) {
                    target = gisSpeedingThreshold;
                    break;
                } else if (vehicleInfoDTO.getProjectId().equals(gisSpeedingThreshold.getProjectId())){
                    target = gisSpeedingThreshold;
                    break;
                } else if ("0".equals(gisSpeedingThreshold.getThresholdType())){
                    target = gisSpeedingThreshold;
                }
            }
        }
        if (target == null) {
            target = new GisSpeedingThreshold();
            target.setIsSpeed(speed.doubleValue() >= standSpeed);
            target.setThresholdName("全局限速");
        } else {
            target.setIsSpeed(speed.doubleValue() >= target.getThresholdSpeed());
        }
        return target;
    }

    //构建Alarm
    public Alarm buildBean(GpsData gpsData, VehicleInfoDTO vehicleInfoDTO, SpeedRedisVO speedRedis, Integer flag, GisSpeedingThreshold speedingThreshold){
        Alarm alarm = new Alarm();
        alarm.setType(2);
        alarm.setFlag(flag);
        alarm.setTime(gpsData.getGpstime());
        DriverSignInDto driverDto =  vehicleBindDriverService.getCacheDriverInfo(vehicleInfoDTO.getPlateNumber());
        GisSpeedingAlarm speedAlarm = new GisSpeedingAlarm();
        speedAlarm.setPlateNumber(vehicleInfoDTO.getPlateNumber());
        speedAlarm.setProjectId(vehicleInfoDTO.getProjectId());
        speedAlarm.setProjectName(vehicleInfoDTO.getProjectName());
        speedAlarm.setGpsNo(vehicleInfoDTO.getLocationTerminalId());
        speedAlarm.setRuleSpeed(BigDecimal.valueOf(speedingThreshold.getThresholdSpeed()));
        speedAlarm.setThresholdName(speedingThreshold.getThresholdName());
        if(Objects.nonNull(driverDto)){
            speedAlarm.setDriverId(driverDto.getDriverId());
            speedAlarm.setDriverName(driverDto.getDriverName());
        }
        if(flag == 1){
            //开始
            speedAlarm.setSpeed(gpsData.getSpeed().setScale(2,BigDecimal.ROUND_HALF_UP));
            //转换高德信息
            LatLonPoint point = CoordinateConverterUtil.toWGSPoint(gpsData.getLat().doubleValue(),gpsData.getLon().doubleValue());
            speedAlarm.setStartLon(new BigDecimal(point.getLongitude()));
            speedAlarm.setStartLat(new BigDecimal(point.getLatitude()));
            speedAlarm.setStartTime(gpsData.getGpstime());
            alarm.setData(com.alibaba.fastjson.JSONObject.parseObject(com.alibaba.fastjson.JSONObject.toJSONString(speedAlarm)));
            return alarm;
        }else if(flag == 2){
            //结束
            //最高车速
            List<Double> speedList = speedRedis.getSpeedList();
            speedAlarm.setSpeed(BigDecimal.valueOf(Collections.max(speedList)));
            //补充最新一次速度，计算平均车速
            speedList.add(gpsData.getSpeed().doubleValue());
            speedAlarm.setAverageSpeed(BigDecimal.valueOf(speedList.stream().mapToDouble(Double::doubleValue).average().orElse(0)));
            LatLonPoint point = CoordinateConverterUtil.toWGSPoint(gpsData.getLat().doubleValue(),gpsData.getLon().doubleValue());
            speedAlarm.setEndLon(new BigDecimal(point.getLongitude()));
            speedAlarm.setEndLat(new BigDecimal(point.getLatitude()));
            speedAlarm.setEndTime(gpsData.getGpstime());
            alarm.setData(com.alibaba.fastjson.JSONObject.parseObject(com.alibaba.fastjson.JSONObject.toJSONString(speedAlarm)));
            return alarm;
        }
        return null;
    }
}
